
"""
A system log management tool.
The main function is to classify log messages based
on their log templates (formats and variable locations),
and to store the data in a database.
"""

__version__ = '0.2.2'

